"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import sys

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions
from builtins import (
    bool,
    bytes,
    int,
    str,
    type,
)
from collections.abc import (
    Iterable,
)
from google.protobuf.descriptor import (
    Descriptor,
    EnumDescriptor,
    FileDescriptor,
)
from google.protobuf.internal.containers import (
    RepeatedCompositeFieldContainer,
    RepeatedScalarFieldContainer,
)
from google.protobuf.internal.enum_type_wrapper import (
    _EnumTypeWrapper,
)
from google.protobuf.message import (
    Message,
)
from typing import (
    NewType,
)

DESCRIPTOR: FileDescriptor

@typing_extensions.final
class HeaderBlock(Message):
    """Contains the file header."""

    DESCRIPTOR: Descriptor

    BBOX_FIELD_NUMBER: int
    REQUIRED_FEATURES_FIELD_NUMBER: int
    OPTIONAL_FEATURES_FIELD_NUMBER: int
    WRITINGPROGRAM_FIELD_NUMBER: int
    SOURCE_FIELD_NUMBER: int
    OSMOSIS_REPLICATION_TIMESTAMP_FIELD_NUMBER: int
    OSMOSIS_REPLICATION_SEQUENCE_NUMBER_FIELD_NUMBER: int
    OSMOSIS_REPLICATION_BASE_URL_FIELD_NUMBER: int
    @property
    def bbox(self) -> HeaderBBox: ...
    @property
    def required_features(self) -> RepeatedScalarFieldContainer[str]:
        """Additional tags to aid in parsing this dataset"""
    @property
    def optional_features(self) -> RepeatedScalarFieldContainer[str]: ...
    writingprogram: str
    source: str
    """From the bbox field."""
    osmosis_replication_timestamp: int
    """Tags that allow continuing an Osmosis replication

    replication timestamp, expressed in seconds since the epoch,
    otherwise the same value as in the "timestamp=..." field
    in the state.txt file used by Osmosis
    """
    osmosis_replication_sequence_number: int
    """replication sequence number (sequenceNumber in state.txt)"""
    osmosis_replication_base_url: str
    """replication base URL (from Osmosis' configuration.txt file)"""
    def __init__(
        self,
        *,
        bbox: HeaderBBox | None = ...,
        required_features: Iterable[str] | None = ...,
        optional_features: Iterable[str] | None = ...,
        writingprogram: str | None = ...,
        source: str | None = ...,
        osmosis_replication_timestamp: int | None = ...,
        osmosis_replication_sequence_number: int | None = ...,
        osmosis_replication_base_url: str | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "bbox",
            b"bbox",
            "osmosis_replication_base_url",
            b"osmosis_replication_base_url",
            "osmosis_replication_sequence_number",
            b"osmosis_replication_sequence_number",
            "osmosis_replication_timestamp",
            b"osmosis_replication_timestamp",
            "source",
            b"source",
            "writingprogram",
            b"writingprogram",
        ],
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "bbox",
            b"bbox",
            "optional_features",
            b"optional_features",
            "osmosis_replication_base_url",
            b"osmosis_replication_base_url",
            "osmosis_replication_sequence_number",
            b"osmosis_replication_sequence_number",
            "osmosis_replication_timestamp",
            b"osmosis_replication_timestamp",
            "required_features",
            b"required_features",
            "source",
            b"source",
            "writingprogram",
            b"writingprogram",
        ],
    ) -> None: ...

@typing_extensions.final
class HeaderBBox(Message):
    """* The bounding box field in the OSM header. BBOX, as used in the OSM
    header. Units are always in nanodegrees -- they do not obey
    granularity rules.
    """

    DESCRIPTOR: Descriptor

    LEFT_FIELD_NUMBER: int
    RIGHT_FIELD_NUMBER: int
    TOP_FIELD_NUMBER: int
    BOTTOM_FIELD_NUMBER: int
    left: int
    right: int
    top: int
    bottom: int
    def __init__(
        self,
        *,
        left: int | None = ...,
        right: int | None = ...,
        top: int | None = ...,
        bottom: int | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "bottom", b"bottom", "left", b"left", "right", b"right", "top", b"top"
        ],
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "bottom", b"bottom", "left", b"left", "right", b"right", "top", b"top"
        ],
    ) -> None: ...

@typing_extensions.final
class PrimitiveBlock(Message):
    """/////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////
    """

    DESCRIPTOR: Descriptor

    STRINGTABLE_FIELD_NUMBER: int
    PRIMITIVEGROUP_FIELD_NUMBER: int
    GRANULARITY_FIELD_NUMBER: int
    LAT_OFFSET_FIELD_NUMBER: int
    LON_OFFSET_FIELD_NUMBER: int
    DATE_GRANULARITY_FIELD_NUMBER: int
    @property
    def stringtable(self) -> StringTable: ...
    @property
    def primitivegroup(self) -> RepeatedCompositeFieldContainer[PrimitiveGroup]: ...
    granularity: int
    """Granularity, units of nanodegrees, used to store coordinates in this block"""
    lat_offset: int
    """Offset value between the output coordinates coordinates and the granularity grid in unites of nanodegrees."""
    lon_offset: int
    date_granularity: int
    """Granularity of dates, normally represented in units of milliseconds since the 1970 epoch."""
    def __init__(
        self,
        *,
        stringtable: StringTable | None = ...,
        primitivegroup: Iterable[PrimitiveGroup] | None = ...,
        granularity: int | None = ...,
        lat_offset: int | None = ...,
        lon_offset: int | None = ...,
        date_granularity: int | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "date_granularity",
            b"date_granularity",
            "granularity",
            b"granularity",
            "lat_offset",
            b"lat_offset",
            "lon_offset",
            b"lon_offset",
            "stringtable",
            b"stringtable",
        ],
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "date_granularity",
            b"date_granularity",
            "granularity",
            b"granularity",
            "lat_offset",
            b"lat_offset",
            "lon_offset",
            b"lon_offset",
            "primitivegroup",
            b"primitivegroup",
            "stringtable",
            b"stringtable",
        ],
    ) -> None: ...

@typing_extensions.final
class PrimitiveGroup(Message):
    """Group of OSMPrimitives. All primitives in a group must be the same type."""

    DESCRIPTOR: Descriptor

    NODES_FIELD_NUMBER: int
    DENSE_FIELD_NUMBER: int
    WAYS_FIELD_NUMBER: int
    RELATIONS_FIELD_NUMBER: int
    CHANGESETS_FIELD_NUMBER: int
    @property
    def nodes(self) -> RepeatedCompositeFieldContainer[Node]: ...
    @property
    def dense(self) -> DenseNodes: ...
    @property
    def ways(self) -> RepeatedCompositeFieldContainer[Way]: ...
    @property
    def relations(self) -> RepeatedCompositeFieldContainer[Relation]: ...
    @property
    def changesets(self) -> RepeatedCompositeFieldContainer[ChangeSet]: ...
    def __init__(
        self,
        *,
        nodes: Iterable[Node] | None = ...,
        dense: DenseNodes | None = ...,
        ways: Iterable[Way] | None = ...,
        relations: Iterable[Relation] | None = ...,
        changesets: Iterable[ChangeSet] | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["dense", b"dense"]
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "changesets",
            b"changesets",
            "dense",
            b"dense",
            "nodes",
            b"nodes",
            "relations",
            b"relations",
            "ways",
            b"ways",
        ],
    ) -> None: ...

@typing_extensions.final
class StringTable(Message):
    """* String table, contains the common strings in each block.

    Note that we reserve index '0' as a delimiter, so the entry at that
    index in the table is ALWAYS blank and unused.
    """

    DESCRIPTOR: Descriptor

    S_FIELD_NUMBER: int
    @property
    def s(self) -> RepeatedScalarFieldContainer[bytes]: ...
    def __init__(
        self,
        *,
        s: Iterable[bytes] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["s", b"s"]) -> None: ...

@typing_extensions.final
class Info(Message):
    """Optional metadata that may be included into each primitive."""

    DESCRIPTOR: Descriptor

    VERSION_FIELD_NUMBER: int
    TIMESTAMP_FIELD_NUMBER: int
    CHANGESET_FIELD_NUMBER: int
    UID_FIELD_NUMBER: int
    USER_SID_FIELD_NUMBER: int
    VISIBLE_FIELD_NUMBER: int
    version: int
    timestamp: int
    changeset: int
    uid: int
    user_sid: int
    """String IDs"""
    visible: bool
    """The visible flag is used to store history information. It indicates that
    the current object version has been created by a delete operation on the
    OSM API.
    When a writer sets this flag, it MUST add a required_features tag with
    value "HistoricalInformation" to the HeaderBlock.
    If this flag is not available for some object it MUST be assumed to be
    true if the file has the required_features tag "HistoricalInformation"
    set.
    """
    def __init__(
        self,
        *,
        version: int | None = ...,
        timestamp: int | None = ...,
        changeset: int | None = ...,
        uid: int | None = ...,
        user_sid: int | None = ...,
        visible: bool | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "changeset",
            b"changeset",
            "timestamp",
            b"timestamp",
            "uid",
            b"uid",
            "user_sid",
            b"user_sid",
            "version",
            b"version",
            "visible",
            b"visible",
        ],
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "changeset",
            b"changeset",
            "timestamp",
            b"timestamp",
            "uid",
            b"uid",
            "user_sid",
            b"user_sid",
            "version",
            b"version",
            "visible",
            b"visible",
        ],
    ) -> None: ...

@typing_extensions.final
class DenseInfo(Message):
    """* Optional metadata that may be included into each primitive. Special dense format used in DenseNodes."""

    DESCRIPTOR: Descriptor

    VERSION_FIELD_NUMBER: int
    TIMESTAMP_FIELD_NUMBER: int
    CHANGESET_FIELD_NUMBER: int
    UID_FIELD_NUMBER: int
    USER_SID_FIELD_NUMBER: int
    VISIBLE_FIELD_NUMBER: int
    @property
    def version(self) -> RepeatedScalarFieldContainer[int]: ...
    @property
    def timestamp(self) -> RepeatedScalarFieldContainer[int]:
        """DELTA coded"""
    @property
    def changeset(self) -> RepeatedScalarFieldContainer[int]:
        """DELTA coded"""
    @property
    def uid(self) -> RepeatedScalarFieldContainer[int]:
        """DELTA coded"""
    @property
    def user_sid(self) -> RepeatedScalarFieldContainer[int]:
        """String IDs for usernames. DELTA coded"""
    @property
    def visible(self) -> RepeatedScalarFieldContainer[bool]:
        """The visible flag is used to store history information. It indicates that
        the current object version has been created by a delete operation on the
        OSM API.
        When a writer sets this flag, it MUST add a required_features tag with
        value "HistoricalInformation" to the HeaderBlock.
        If this flag is not available for some object it MUST be assumed to be
        true if the file has the required_features tag "HistoricalInformation"
        set.
        """
    def __init__(
        self,
        *,
        version: Iterable[int] | None = ...,
        timestamp: Iterable[int] | None = ...,
        changeset: Iterable[int] | None = ...,
        uid: Iterable[int] | None = ...,
        user_sid: Iterable[int] | None = ...,
        visible: Iterable[bool] | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "changeset",
            b"changeset",
            "timestamp",
            b"timestamp",
            "uid",
            b"uid",
            "user_sid",
            b"user_sid",
            "version",
            b"version",
            "visible",
            b"visible",
        ],
    ) -> None: ...

@typing_extensions.final
class ChangeSet(Message):
    """THIS IS STUB DESIGN FOR CHANGESETS. NOT USED RIGHT NOW.
    TODO:    REMOVE THIS?
    """

    DESCRIPTOR: Descriptor

    ID_FIELD_NUMBER: int
    id: int
    """
      // Parallel arrays.
      repeated uint32 keys = 2 [packed = true]; // String IDs.
      repeated uint32 vals = 3 [packed = true]; // String IDs.

      optional Info info = 4;
    """
    def __init__(
        self,
        *,
        id: int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["id", b"id"]) -> bool: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["id", b"id"]
    ) -> None: ...

@typing_extensions.final
class Node(Message):
    DESCRIPTOR: Descriptor

    ID_FIELD_NUMBER: int
    KEYS_FIELD_NUMBER: int
    VALS_FIELD_NUMBER: int
    INFO_FIELD_NUMBER: int
    LAT_FIELD_NUMBER: int
    LON_FIELD_NUMBER: int
    id: int
    @property
    def keys(self) -> RepeatedScalarFieldContainer[int]:
        """Parallel arrays.
        String IDs.
        """
    @property
    def vals(self) -> RepeatedScalarFieldContainer[int]:
        """String IDs."""
    @property
    def info(self) -> Info:
        """May be omitted in omitmeta"""
    lat: int
    lon: int
    def __init__(
        self,
        *,
        id: int | None = ...,
        keys: Iterable[int] | None = ...,
        vals: Iterable[int] | None = ...,
        info: Info | None = ...,
        lat: int | None = ...,
        lon: int | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "id", b"id", "info", b"info", "lat", b"lat", "lon", b"lon"
        ],
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "id",
            b"id",
            "info",
            b"info",
            "keys",
            b"keys",
            "lat",
            b"lat",
            "lon",
            b"lon",
            "vals",
            b"vals",
        ],
    ) -> None: ...

@typing_extensions.final
class DenseNodes(Message):
    """Used to densly represent a sequence of nodes that do not have any tags.

    We represent these nodes columnwise as five columns: ID's, lats, and
    lons, all delta coded. When metadata is not omitted,

    We encode keys & vals for all nodes as a single array of integers
    containing key-stringid and val-stringid, using a stringid of 0 as a
    delimiter between nodes.

    ( (<keyid> <valid>)* '0' )*
    """

    DESCRIPTOR: Descriptor

    ID_FIELD_NUMBER: int
    DENSEINFO_FIELD_NUMBER: int
    LAT_FIELD_NUMBER: int
    LON_FIELD_NUMBER: int
    KEYS_VALS_FIELD_NUMBER: int
    @property
    def id(self) -> RepeatedScalarFieldContainer[int]:
        """DELTA coded"""
    @property
    def denseinfo(self) -> DenseInfo:
        """repeated Info info = 4;"""
    @property
    def lat(self) -> RepeatedScalarFieldContainer[int]:
        """DELTA coded"""
    @property
    def lon(self) -> RepeatedScalarFieldContainer[int]:
        """DELTA coded"""
    @property
    def keys_vals(self) -> RepeatedScalarFieldContainer[int]:
        """Special packing of keys and vals into one array. May be empty if all nodes in this block are tagless."""
    def __init__(
        self,
        *,
        id: Iterable[int] | None = ...,
        denseinfo: DenseInfo | None = ...,
        lat: Iterable[int] | None = ...,
        lon: Iterable[int] | None = ...,
        keys_vals: Iterable[int] | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["denseinfo", b"denseinfo"]
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "denseinfo",
            b"denseinfo",
            "id",
            b"id",
            "keys_vals",
            b"keys_vals",
            "lat",
            b"lat",
            "lon",
            b"lon",
        ],
    ) -> None: ...

@typing_extensions.final
class Way(Message):
    DESCRIPTOR: Descriptor

    ID_FIELD_NUMBER: int
    KEYS_FIELD_NUMBER: int
    VALS_FIELD_NUMBER: int
    INFO_FIELD_NUMBER: int
    REFS_FIELD_NUMBER: int
    id: int
    @property
    def keys(self) -> RepeatedScalarFieldContainer[int]:
        """Parallel arrays."""
    @property
    def vals(self) -> RepeatedScalarFieldContainer[int]: ...
    @property
    def info(self) -> Info: ...
    @property
    def refs(self) -> RepeatedScalarFieldContainer[int]:
        """DELTA coded"""
    def __init__(
        self,
        *,
        id: int | None = ...,
        keys: Iterable[int] | None = ...,
        vals: Iterable[int] | None = ...,
        info: Info | None = ...,
        refs: Iterable[int] | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["id", b"id", "info", b"info"]
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "id",
            b"id",
            "info",
            b"info",
            "keys",
            b"keys",
            "refs",
            b"refs",
            "vals",
            b"vals",
        ],
    ) -> None: ...

@typing_extensions.final
class Relation(Message):
    DESCRIPTOR: Descriptor

    class _MemberType:
        ValueType = NewType("ValueType", int)
        V: typing_extensions.TypeAlias = ValueType

    class _MemberTypeEnumTypeWrapper(
        _EnumTypeWrapper[Relation._MemberType.ValueType], type
    ):  # noqa: F821
        DESCRIPTOR: EnumDescriptor
        NODE: Relation._MemberType.ValueType  # 0
        WAY: Relation._MemberType.ValueType  # 1
        RELATION: Relation._MemberType.ValueType  # 2

    class MemberType(_MemberType, metaclass=_MemberTypeEnumTypeWrapper): ...
    NODE: Relation.MemberType.ValueType  # 0
    WAY: Relation.MemberType.ValueType  # 1
    RELATION: Relation.MemberType.ValueType  # 2

    ID_FIELD_NUMBER: int
    KEYS_FIELD_NUMBER: int
    VALS_FIELD_NUMBER: int
    INFO_FIELD_NUMBER: int
    ROLES_SID_FIELD_NUMBER: int
    MEMIDS_FIELD_NUMBER: int
    TYPES_FIELD_NUMBER: int
    id: int
    @property
    def keys(self) -> RepeatedScalarFieldContainer[int]:
        """Parallel arrays."""
    @property
    def vals(self) -> RepeatedScalarFieldContainer[int]: ...
    @property
    def info(self) -> Info: ...
    @property
    def roles_sid(self) -> RepeatedScalarFieldContainer[int]:
        """Parallel arrays
        This should have been defined as uint32 for consistency, but it is now too late to change it
        """
    @property
    def memids(self) -> RepeatedScalarFieldContainer[int]:
        """DELTA encoded"""
    @property
    def types(self) -> RepeatedScalarFieldContainer[Relation.MemberType.ValueType]: ...
    def __init__(
        self,
        *,
        id: int | None = ...,
        keys: Iterable[int] | None = ...,
        vals: Iterable[int] | None = ...,
        info: Info | None = ...,
        roles_sid: Iterable[int] | None = ...,
        memids: Iterable[int] | None = ...,
        types: Iterable[Relation.MemberType.ValueType] | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["id", b"id", "info", b"info"]
    ) -> bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "id",
            b"id",
            "info",
            b"info",
            "keys",
            b"keys",
            "memids",
            b"memids",
            "roles_sid",
            b"roles_sid",
            "types",
            b"types",
            "vals",
            b"vals",
        ],
    ) -> None: ...



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Generator, TypeAlias

import geopandas as gpd
import pandas as pd
from shapely import Polygon

from .blocks import read_blocks
from .filter import filter_groups
from .nodes import consolidate_nodes
from .proto import HeaderBlock, PrimitiveBlock
from .relations import consolidate_relations
from .unpacked import BaseGroup, NodesGroup, RelationGroup, WayGroup
from .ways import consolidate_ways

__all__ = ["OSMFile"]

BBox: TypeAlias = tuple[float, float, float, float] | Polygon
ReferenceDict: TypeAlias = defaultdict[str, set[int]]


def _unpack_primitive_block(
    block: PrimitiveBlock,
) -> Generator[BaseGroup, None, None]:
    string_table: list[str] = [x.decode("utf-8") for x in block.stringtable.s]

    for group in block.primitivegroup:
        # each group contains only one field at a time, where the fields can be
        # nodes, dense, ways, relations or changesets

        if len(group.nodes) > 0:
            raise NotImplementedError()

        if len(group.dense.id) > 0:
            yield NodesGroup.from_dense_group(
                group,
                string_table,
                granularity=block.granularity,
                lat_offset=block.lat_offset,
                lon_offset=block.lon_offset,
            )

        if len(group.ways) > 0:
            yield WayGroup.from_primitive_group(group, string_table)

        if len(group.relations) > 0:
            yield RelationGroup.from_primitive_group(group, string_table)


def _read_and_unpack_groups(
    file_iterator: Generator[bytes, None, None],
) -> Generator[BaseGroup, None, None]:
    """Parse all nodes from a file-iterator."""
    for block in file_iterator:
        yield from _unpack_primitive_block(PrimitiveBlock.FromString(block))


@dataclass
class OSMFile:
    nodes: list[NodesGroup] = field(default_factory=list)
    ways: list[WayGroup] = field(default_factory=list)
    relations: list[RelationGroup] = field(default_factory=list)

    @classmethod
    def from_file(cls, fp: Path | str) -> OSMFile:
        if isinstance(fp, str):
            fp = Path(fp)

        nodes: list[NodesGroup] = []
        ways: list[WayGroup] = []
        relations: list[RelationGroup] = []

        with open(fp, "rb") as f:
            file_iterator = read_blocks(f)

            # fixme: do something with header block here
            _: HeaderBlock = HeaderBlock.FromString(next(file_iterator))

            for group in _read_and_unpack_groups(file_iterator):
                match group:
                    case NodesGroup():
                        nodes.append(group)
                    case WayGroup():
                        ways.append(group)
                    case RelationGroup():
                        relations.append(group)

        return cls(nodes, ways, relations)

    def filter(self, *, tags: set[str]) -> OSMFile:
        self.relations, references = filter_groups(self.relations, tags=tags)
        self.ways, references = filter_groups(
            self.ways, tags=tags, references=references
        )
        self.nodes, _ = filter_groups(self.nodes, tags=tags, references=references)
        return self

    def consolidate(self) -> gpd.GeoDataFrame:
        _node_parts = [
            consolidate_nodes(nodes) for nodes in self.nodes if len(nodes.ids) > 0
        ]
        if len(_node_parts) > 0:
            nodes = pd.concat(_node_parts)
        else:
            raise ValueError("Nothing to consolidate.")

        _way_parts = [
            consolidate_ways(ways, nodes=nodes)
            for ways in self.ways
            if len(ways.ids) > 0
        ]

        if len(_way_parts) > 0:
            ways = pd.concat(_way_parts)
        else:
            ways = gpd.GeoDataFrame()

        _relation_parts = [
            consolidate_relations(relations, ways=ways, nodes=nodes)
            for relations in self.relations
            if len(relations.ids) > 0
        ]

        if len(_relation_parts) > 0:
            relations = pd.concat(_relation_parts)
            return pd.concat([nodes, ways, relations])
        else:
            return pd.concat([nodes, ways])


#  header_bbox = box(
#      header_block.bbox.bottom * 1e-9,
#      header_block.bbox.left * 1e-9,
#      header_block.bbox.top * 1e-9,
#      header_block.bbox.right * 1e-9,
#  )
#
#  if bounding_box is not None and not intersects(header_bbox, bounding_box):



from __future__ import annotations

from dataclasses import dataclass
from itertools import accumulate
from typing import Generator, Iterator, Sequence

import numpy as np
from numpy.typing import NDArray

from osm4gpd.proto import PrimitiveGroup

from .base import BaseGroup

__all__ = ["NodesGroup"]


@dataclass(repr=False)
class NodesGroup(BaseGroup):
    lat: NDArray[np.float64]
    lon: NDArray[np.float64]

    @classmethod
    def from_dense_group(
        cls,
        group: PrimitiveGroup,
        string_table: list[str],
        *,
        granularity: float,
        lat_offset: float,
        lon_offset: float,
    ) -> NodesGroup:
        return cls(
            ids=np.fromiter(accumulate(group.dense.id), dtype=np.int64),
            lat=np.fromiter(
                (
                    (x * granularity + lat_offset) * 1e-9
                    for x in accumulate(group.dense.lat)
                ),
                dtype=np.float64,
            ),
            lon=np.fromiter(
                (
                    (x * granularity + lon_offset) * 1e-9
                    for x in accumulate(group.dense.lon)
                ),
                dtype=np.float64,
            ),
            tags=dict(_parse_dense_tags(group.dense.keys_vals, string_table)),
            version=list(group.dense.denseinfo.version),
            visible=list(_visible(group.dense.denseinfo.visible, len(group.dense.id))),
            changeset=list(accumulate(group.dense.denseinfo.changeset)),
        )


def _parse_dense_tags(
    keys_vals: Sequence[int], string_table: list[str]
) -> Generator[tuple[int, dict[str, str]], None, None]:
    node_idx = 0
    kv_idx = 0

    while kv_idx < len(keys_vals):
        tags = dict()
        while keys_vals[kv_idx] != 0:
            k = keys_vals[kv_idx]
            v = keys_vals[kv_idx + 1]
            kv_idx += 2
            tags[string_table[k]] = string_table[v]

        if len(tags) > 0:
            yield node_idx, tags

        kv_idx += 1
        node_idx += 1


def _visible(values: Sequence[bool], length: int) -> Iterator[bool]:
    if len(values) == length:
        return iter(values)
    elif len(values) == 0:
        return (True for _ in range(length))
    else:
        raise ValueError("Invalid length of 'visible' values")



from typing import TypeAlias, TypeVar

import numpy as np
from numpy.typing import NDArray

from .references import find_references
from .unpacked import BaseGroup, NodesGroup, RelationGroup, WayGroup

ReferenceDict: TypeAlias = dict[str, NDArray[np.int64]]
GroupType = TypeVar("GroupType", bound=BaseGroup)


def get_elements_matching_tags(group: BaseGroup, tags: set[str]) -> NDArray[np.int64]:
    """Return a set of osm ids that matches the given tags."""
    return np.unique(
        np.fromiter(
            (
                group.ids[idx]
                for idx, element_tags in group.tags.items()
                if len(tags.intersection(element_tags.keys())) > 0
            ),
            dtype=np.int64,
        )
    )


def filter_groups(
    groups: list[GroupType], tags: set[str], references: ReferenceDict | None = None
) -> tuple[list[GroupType], ReferenceDict]:
    if references is None:
        references = {}

    if len(groups) == 0:
        return groups, references

    matching_ids: NDArray[np.int64] = np.concatenate(
        list(get_elements_matching_tags(group, tags) for group in groups)
    )

    match groups[0]:
        case RelationGroup():
            for k, v in find_references(
                np.concatenate([matching_ids, references.get("relation", np.array([], dtype=np.int64))]), groups  # type: ignore[type-var]
            ).items():
                references[k] = np.unique(
                    np.concatenate([references.get(k, np.array([], dtype=np.int64)), v])
                )
        case WayGroup():
            for k, v in find_references(
                np.concatenate([matching_ids, references.get("way", np.array([], dtype=np.int64))]), groups  # type: ignore[type-var]
            ).items():
                references[k] = np.unique(
                    np.concatenate([references.get(k, np.array([], dtype=np.int64)), v])
                )
        case _:
            pass

    for group in groups:
        match group:
            case RelationGroup():
                keep = np.where(
                    np.isin(group.ids, matching_ids)
                    | np.isin(group.ids, references.get("relation", []))
                )[0]

                group.ids = group.ids[keep]
                group.version = [group.version[idx] for idx in keep]
                group.member_ids = [group.member_ids[idx] for idx in keep]
                group.member_types = [group.member_types[idx] for idx in keep]
                group.member_roles = [group.member_roles[idx] for idx in keep]
                group.tags = {i: group.tags.get(idx, {}) for i, idx in enumerate(keep)}
                group.visible = [group.visible[idx] for idx in keep]
                group.changeset = [group.changeset[idx] for idx in keep]
            case WayGroup():
                keep = np.where(
                    np.isin(group.ids, matching_ids)
                    | np.isin(group.ids, references.get("way", []))
                )[0]

                group.ids = group.ids[keep]
                group.version = [group.version[idx] for idx in keep]
                group.member_ids = [group.member_ids[idx] for idx in keep]
                group.tags = {i: group.tags.get(idx, {}) for i, idx in enumerate(keep)}
                group.visible = [group.visible[idx] for idx in keep]
                group.changeset = [group.changeset[idx] for idx in keep]
            case NodesGroup():
                keep = np.where(
                    np.isin(group.ids, matching_ids)
                    | np.isin(group.ids, references.get("node", []))
                )[0]
                group.ids = group.ids[keep]
                group.version = [group.version[idx] for idx in keep]
                group.tags = {i: group.tags.get(idx, {}) for i, idx in enumerate(keep)}
                group.visible = [group.visible[idx] for idx in keep]
                group.changeset = [group.changeset[idx] for idx in keep]
                group.lat = group.lat[keep]
                group.lon = group.lon[keep]

    return groups, references



from typing import TypeAlias, TypeVar

import numpy as np
from numpy.typing import NDArray

from .unpacked import RelationGroup, WayGroup

ReferenceDict: TypeAlias = dict[str, NDArray[np.int64]]
GroupType = TypeVar("GroupType", RelationGroup, WayGroup)


def union(ids: list[NDArray[np.int64]]) -> NDArray[np.int64]:
    try:
        return np.unique(np.concatenate(ids))
    except ValueError:
        return np.array([], dtype=np.int64)


def _get_references_from_group(
    group: GroupType, ids: NDArray[np.int64]
) -> ReferenceDict:
    references = {}

    match group:
        case WayGroup():
            references["node"] = union(
                [group.member_ids[idx] for idx in np.where(np.isin(group.ids, ids))[0]]
            )
        case RelationGroup():
            references["node"] = union(
                [
                    group.member_ids[idx][group.member_types[idx] == "node"]
                    for idx in np.where(np.isin(group.ids, ids))[0]
                ]
            )

            references["way"] = union(
                [
                    group.member_ids[idx][group.member_types[idx] == "way"]
                    for idx in np.where(np.isin(group.ids, ids))[0]
                ]
            )

            references["relation"] = union(
                [
                    group.member_ids[idx][group.member_types[idx] == "relation"]
                    for idx in np.where(np.isin(group.ids, ids))[0]
                ]
            )
        case _:
            raise NotImplementedError()

    return references


def find_references(keep: NDArray[np.int64], groups: list[GroupType]) -> ReferenceDict:
    """Recursively find all relation/way/node ids, referenced by relation/way
    ids in the initial set `keep`."""
    references: dict[str, NDArray[np.int64]] = {}

    if len(groups) == 0 or len(keep) == 0:
        return references

    for group in groups:
        for k, v in _get_references_from_group(group, keep).items():
            references[k] = np.unique(
                np.concatenate([references.get(k, np.array([], dtype=np.int64)), v])
            )

    match groups[0]:
        case RelationGroup():
            # Recursively find all other references
            for k, v in find_references(
                np.setdiff1d(np.fromiter(references["relation"], dtype=np.int64), keep),
                groups,
            ).items():
                references[k] = np.unique(
                    np.concatenate([references.get(k, np.array([], dtype=np.int64)), v])
                )
            return references
        case WayGroup():
            return references
        case _:
            raise NotImplementedError()



from __future__ import annotations

from dataclasses import dataclass
from itertools import accumulate

import numpy as np
from numpy.typing import NDArray

from osm4gpd.proto import PrimitiveGroup
from osm4gpd.tags import get_tags

from .base import BaseGroup

__all__ = ["RelationGroup"]


@dataclass(repr=False)
class RelationGroup(BaseGroup):
    member_types: list[NDArray[np.object_]]
    member_roles: list[NDArray[np.object_]]
    member_ids: list[NDArray[np.int64]]

    @classmethod
    def from_primitive_group(
        cls, group: PrimitiveGroup, string_table: list[str]
    ) -> RelationGroup:
        ids: list[int] = []
        versions: list[int] = []
        member_ids: list[NDArray[np.int64]] = []
        member_types: list[NDArray[np.object_]] = []
        member_roles: list[NDArray[np.object_]] = []
        tags: dict[int, dict[str, str]] = {}
        visible: list[bool] = []
        changeset: list[int] = []

        for i, relation in enumerate(group.relations):
            ids.append(relation.id)

            member_types.append(
                np.fromiter(
                    (
                        relation.MemberType.keys()[type_].lower()
                        for type_ in relation.types
                    ),
                    dtype=np.object_,
                )
            )
            member_ids.append(np.fromiter(accumulate(relation.memids), dtype=np.int64))
            member_roles.append(
                np.fromiter(
                    (string_table[sid] for sid in relation.roles_sid), dtype=np.object_
                )
            )

            _tags = get_tags(relation, string_table)
            if len(_tags) > 0:
                tags[i] = _tags

            # fixme: add optional here
            versions.append(relation.info.version)
            visible.append(relation.info.visible)
            changeset.append(relation.info.changeset)

        return cls(
            ids=np.array(ids),
            tags=tags,
            version=versions,
            changeset=changeset,
            visible=visible,
            member_ids=member_ids,
            member_roles=member_roles,
            member_types=member_types,
        )



from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray

__all__ = ["BaseGroup"]


@dataclass(repr=False)
class BaseGroup:
    ids: NDArray[np.int64]
    tags: dict[int, dict[str, str]]
    version: list[int]
    visible: list[bool]
    changeset: list[int]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: fileformat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x10\x66ileformat.proto\x12\x06OSMPBF"l\n\x04\x42lob\x12\x0b\n\x03raw\x18\x01 \x01(\x0c\x12\x10\n\x08raw_size\x18\x02 \x01(\x05\x12\x11\n\tzlib_data\x18\x03 \x01(\x0c\x12\x11\n\tlzma_data\x18\x04 \x01(\x0c\x12\x1f\n\x13OBSOLETE_bzip2_data\x18\x05 \x01(\x0c\x42\x02\x18\x01"?\n\nBlobHeader\x12\x0c\n\x04type\x18\x01 \x02(\t\x12\x11\n\tindexdata\x18\x02 \x01(\x0c\x12\x10\n\x08\x64\x61tasize\x18\x03 \x02(\x05\x42\x11\n\rcrosby.binaryH\x03'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "fileformat_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\rcrosby.binaryH\003"
    _BLOB.fields_by_name["OBSOLETE_bzip2_data"]._options = None
    _BLOB.fields_by_name["OBSOLETE_bzip2_data"]._serialized_options = b"\030\001"
    _globals["_BLOB"]._serialized_start = 28
    _globals["_BLOB"]._serialized_end = 136
    _globals["_BLOBHEADER"]._serialized_start = 138
    _globals["_BLOBHEADER"]._serialized_end = 201
# @@protoc_insertion_point(module_scope)



import logging
from functools import partial
from typing import Generator

import geopandas as gpd
import numpy as np
import pandas as pd
from numpy.typing import NDArray
from shapely import (
    Geometry,
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPolygon,
    Polygon,
    unary_union,
)
from shapely.ops import linemerge

from .unpacked import RelationGroup

logger = logging.getLogger(__name__)


class ConsolidationError(Exception):
    pass


class UnknownInputException(Exception):
    pass


def consolidate_polygons(parts: gpd.GeoSeries) -> Generator[Polygon, None, None]:
    for gtype, geoms in parts.groupby(parts.geom_type):
        match gtype:
            case "Polygon":
                yield from iter(geoms)
            case "LinearRing":
                yield from (Polygon(g) for g in geoms)
            case "LineString":
                merged = linemerge(iter(geoms))

                if isinstance(merged, LineString):
                    yield Polygon(merged.coords)
                else:
                    yield from (Polygon(g.coords) for g in merged.geoms)
            case _:
                raise ConsolidationError(
                    f"Could not consolidate Polygon for type {gtype}"
                )


def consolidate_linestrings(parts: gpd.GeoSeries) -> Generator[LineString, None, None]:
    for gtype, geoms in parts.groupby(parts.geom_type):
        match gtype:
            case "Polygon":
                yield from (poly.boundary for poly in geoms)
            case "LinearRing" | "LineString":
                yield from iter(geoms)
            case _:
                raise ConsolidationError(
                    f"Could not consolidate LineString for type {gtype}"
                )


def parse_multipolygon_relation(
    members: NDArray[np.int64],
    roles: NDArray[np.object_],
    types: NDArray[np.object_],
    ways: gpd.GeoDataFrame,
    relations: gpd.GeoDataFrame | None = None,
) -> Polygon | MultiPolygon:
    if (types == "relation").any() and relations is None:
        raise ConsolidationError(
            "Can not consolidate multipolygon that depends on relations."
        )
    elif (types == "relation").any() and relations is not None:
        outer_ = relations.loc[
            members[(types == "relation") & (roles == "outer")], "geometry"
        ].to_list()
        inner_ = relations.loc[
            members[(types == "relation") & (roles == "inner_")], "geometry"
        ].to_list()
    else:
        outer_ = []
        inner_ = []

    loc = np.isin(members, ways.index)
    roles = roles[loc]
    geoms = ways.loc[members[loc], "geometry"]

    outer = list(consolidate_polygons(geoms[roles == "outer"]))
    inner = list(consolidate_polygons(geoms[roles == "inner"]))

    return unary_union(outer + outer_).difference(unary_union(inner + inner_))


def parse_boundary_relation(
    members: NDArray[np.int64],
    roles: NDArray[np.object_],
    types: NDArray[np.object_],
    ways: gpd.GeoDataFrame,
) -> LineString | MultiLineString:
    """Boundaries are parsed as LineString, 'label', 'admin_centre' and
    'subarea' roles will be ignored.

    For more information, see
    https://wiki.openstreetmap.org/wiki/Relation:boundary#Relation_members
    """
    idx = (
        np.isin(members, ways.index)
        & (types == "way")
        & np.isin(roles, ("outer", "inner"))
    )
    return linemerge(consolidate_linestrings(ways.loc[members[idx], "geometry"]))


def parse_generic_relation(
    members: NDArray[np.int64],
    types: NDArray[np.object_],
    *,
    ways: gpd.GeoDataFrame,
    nodes: gpd.GeoDataFrame,
    relations: gpd.GeoDataFrame | None = None,
) -> GeometryCollection:
    geoms = (
        ways.loc[ways.index.intersection(members[types == "way"]), "geometry"].to_list()
        + nodes.loc[
            nodes.index.intersection(members[types == "node"]), "geometry"
        ].to_list()
    )

    if relations is not None:
        geoms += relations.loc[
            relations.index.intersection(members[types == "relation"]), "geometry"
        ].to_list()

    return GeometryCollection(geoms)


def _consolidate_geometries(
    group: RelationGroup, ways: gpd.GeoDataFrame, nodes: gpd.GeoDataFrame
) -> Generator[tuple[int, Geometry | partial], None, None]:
    for idx, (members, roles, types) in enumerate(
        zip(group.member_ids, group.member_roles, group.member_types)
    ):
        match group.tags.get(idx, {"type": "generic"})["type"]:
            case "multipolygon":
                if (types == "relation").any():
                    yield idx, partial(
                        parse_multipolygon_relation,
                        members=members,
                        roles=roles,
                        types=types,
                        ways=ways,
                    )
                    continue

                yield idx, parse_multipolygon_relation(members, roles, types, ways)
            case "boundary":
                yield idx, parse_boundary_relation(members, roles, types, ways)
            case _:
                if (types == "relation").any():
                    yield idx, partial(
                        parse_generic_relation,
                        members=members,
                        types=types,
                        ways=ways,
                        nodes=nodes,
                    )
                    continue

                yield idx, parse_generic_relation(
                    members, types, ways=ways, nodes=nodes
                )


def consolidate_relations(
    group: RelationGroup, ways: gpd.GeoDataFrame, nodes: gpd.GeoDataFrame
) -> gpd.GeoDataFrame:
    resolved_geometries: list[Geometry] = []
    resolved_idx: list[int] = []
    unresolved: list[tuple[int, Geometry]] = []

    for idx, geometry in _consolidate_geometries(group, ways, nodes):
        match geometry:
            case Geometry():  # type: ignore[misc]
                resolved_geometries.append(geometry)
                resolved_idx.append(idx)
            case partial():
                unresolved.append((idx, geometry))

    relations = gpd.GeoDataFrame(
        {
            "idx": resolved_idx,
            "id": group.ids[resolved_idx],
            "geometry": resolved_geometries,
            "version": np.array(group.version)[resolved_idx],
            "changeset": np.array(group.changeset)[resolved_idx],
            "visible": np.array(group.visible)[resolved_idx],
        },
        crs="EPSG:4326",
    )

    if len(unresolved) > 0:
        relations = pd.concat(
            [
                relations,
                gpd.GeoDataFrame.from_dict(
                    {
                        idx: {
                            "geometry": func(relations=relations.set_index("id")),
                            "version": group.version[idx],
                            "changeset": group.version[idx],
                            "visible": group.version[idx],
                            "id": group.ids[idx],
                        }
                        for idx, func in unresolved
                    },
                    orient="index",
                    crs="EPSG:4326",
                ),
            ]
        )

    return relations.join(
        pd.DataFrame.from_dict(
            group.tags,
            orient="index",
            dtype=str(pd.SparseDtype(str)),
        )
    ).set_index("id")



from typing import Generator, Sequence

from .proto import Node, Relation, Way


def get_tags(obj: Relation | Way | Node, string_table: list[str]) -> dict[str, str]:
    return {string_table[k]: string_table[v] for k, v in zip(obj.keys, obj.vals)}


def parse_dense_tags(
    keys_vals: Sequence[int], string_table: list[str]
) -> Generator[tuple[int, dict[str, str]], None, None]:
    node_idx = 0
    kv_idx = 0

    while kv_idx < len(keys_vals):
        tags = dict()
        while keys_vals[kv_idx] != 0:
            k = keys_vals[kv_idx]
            v = keys_vals[kv_idx + 1]
            kv_idx += 2
            tags[string_table[k]] = string_table[v]

        if len(tags) > 0:
            yield node_idx, tags

        kv_idx += 1
        node_idx += 1
